import PythonQt
import urllib3
import json
import TokenManager
from TransporterModelsWebService import TransporterModelsWebService
import PhenarisLogger
import app
import time
from SdfHelper import getSDFileTextFromUrl

token = None
token_set = False
log = PhenarisLogger.PhenarisQueryLogger('PhenarisQuery.log')

BASE_AUTH_URL = 'https://178.115.234.79:10443/auth/'  # TransporterModels authentication URL
model_colors = {
    'BCRP': (128, 0, 0),
    'BSEP': (255, 0, 0),
    'MATE1': (0, 128, 0),
    'MRP3': (0, 255, 0),
    'OATP1B1': (0, 0, 128),
    'OATP1B3': (0, 0, 255),
    'OCT1': (128, 0, 128),
    'OCT2': (128, 0, 255),
    'P-gp (MDR1)': (255, 0, 128)
}


def _get_token(appui):
    """
    Receives a token and a refresh token from Transporter Models and stores it in the TokenManager object for the session.
    :param appui:
    :return: Returns a TokenManager Object if the login process was successful
    """
    global token
    if not token:
        token = _login_dialog(appui)
    else:
        response = _token_refresh(token)
        if response.status == 200:
            token_str = json.loads(response.data.decode('utf-8'))['access_token']
            token = TokenManager.TokenManager(token_str, token.getRefreshToken())
        else:
            appui.show("You got automatically logged out!")
            token = _login_dialog(appui)
    return token


def _create_web_service(appui):
    """
    Creates an instance of a web service with the token from the TokenManager
    to communicate with the Transporter Models API.

    If there is no token the function will throw a ValueError.
    :param appui:
    :return: Instance of the TransporterModelsWebService
    """
    try:
        access_token = _get_token(appui)
        if access_token is None:
            web_service = None
        else:
            web_service = TransporterModelsWebService(access_token)
    except Exception as e:
        raise Exception(e)
    return web_service


def _login_dialog(appui):
    dlg = PythonQt.LogInDialog()
    dlg.setWindowTitle("Transporter Models Login")
    if dlg.run():
        try:
            username = dlg.getUser()
            password = dlg.getPassword()
            if username == "" or password == "":
                appui.show("Username and password must not be empty.")
            else:
                # send login data to api and get token
                http = urllib3.PoolManager(cert_reqs='CERT_NONE')
                url = BASE_AUTH_URL + "login"
                data = {'customername': 'BetaTesters',
                        'username': username,
                        'password': password}
                encoded_data = json.dumps(data).encode('utf-8')
                response = http.request('POST', url, body=encoded_data,
                                        headers={'Content-Type': 'application/json'})
                if response.status == 200:
                    token_str = json.loads(response.data.decode('utf-8'))['access_token']
                    refresh_token_str = json.loads(response.data.decode('utf-8'))['refresh_token']
                    token = TokenManager.TokenManager(token_str, refresh_token_str)
                    return token
                elif response.status == 403:
                    appui.show("Username or password is wrong.")
                else:
                    log.log("Something went wrong: " + str(response.status) + " " + response.reason)
                    appui.show("Something went wrong: " + str(response.status) + " " + response.reason)

        except Exception as e:
            raise Exception(e)
    else:
        return None


def _token_refresh(token):
    # Get refresh token every time when call is made
    http = urllib3.PoolManager(cert_reqs='CERT_NONE')
    url = BASE_AUTH_URL + "refresh"
    headers = {'Authorization': 'Bearer ' + token.getRefreshToken()}
    response = http.request('POST', url,
                            headers=headers)
    return response


def _create_prediction_columns(id_response_list):
    """
    Create columns for the "Prediction" and "Submit SDF file" dataset in StarDrop
    :return: cols List of Dataset column types
    """

    cols = list()

    model_header_dict = {}  # Dictionary for the model names
    domain_header_dict = {}  # Dictionary for the domain names
    # Insert the name of the models in column list
    for i in range(0, len(id_response_list)):
        for j in range(0, len(id_response_list[i]['result'])):
            # id_response_keys = list(id_response_list[i]['result'][j]['res'].keys())
            id_response_keys = _get_response_result_keys(id_response_list[i]['result'][j])
            for key in id_response_keys:
                # model_header_dict[key] = app.Dataset.meta(key, app.cont, ((0.0, 1.0), (0, 0)))
                cat_header = app.Dataset.createCategoryHeader(key, ["yes", "no"])
                model_header_dict[key] = cat_header[0]

    # Insert the name of the applicability domain in column list
    for i in range(0, len(id_response_list)):
        for j in range(0, len(list(id_response_list[i]['result']))):
            for k in range(0, len(list(id_response_list[i]['result'][j]['ad']))):
                name = str(id_response_list[i]['result'][j]['ad'][k]['model']) + " - " + str(
                    id_response_list[i]['result'][j]['ad'][k]['method'])
                domain_header_dict[name] = app.Dataset.meta(name)

    # Sorts the model and domain keys alphabetically and by classifier in case of models and insert them to "cols"
    sorted_model_keys = sort_model_keys(model_header_dict.keys())
    for key in sorted_model_keys:
        cols.append(model_header_dict[key])
    sorted_domain_keys = sorted(domain_header_dict.keys())
    for key in sorted_domain_keys:
        cols.append(domain_header_dict[key])

    return cols


def _create_single_molecule_prediction_columns(response_result_keys, response_result_domain):
    """
    Create columns for the "Single prediction using selected model" dataset in StarDrop
    :return: cols List of Dataset column types
    """

    cols = list()

    model_header_dict = {}  # Dictionary for the model names
    domain_header_dict = {}  # Dictionary for the domain names
    # Insert the name of the models in column list

    for key in response_result_keys:
        # model_header_dict[key] = app.Dataset.meta(key, app.cont, ((0.0, 1.0), (0, 0)))
        cat_header = app.Dataset.createCategoryHeader(key, ["yes", "no"])
        model_header_dict[key] = cat_header[0]

    for i in range(0, len(response_result_domain)):
        name = str(response_result_domain[i]['model'] + " - " + response_result_domain[i]['method'])
        domain_header_dict[name] = app.Dataset.meta(name)

    # Sorts the model and domain keys alphabetically and by classifier in case of models and insert them to "cols"
    sorted_model_keys = sort_model_keys(model_header_dict.keys())
    for key in sorted_model_keys:
        cols.append(model_header_dict[key])
    sorted_domain_keys = sorted(domain_header_dict.keys())
    for key in sorted_domain_keys:
        cols.append(domain_header_dict[key])

    return cols


def _create_single_prediction_columns(response_result_keys, response_result_domain):
    """
    Create columns for "Single Prediction" dataset in StarDrop
    :return: cols List of Dataset column types
    """

    cols = list()
    cols.append(app.Dataset.meta('ID'))
    cols.append(app.Dataset.meta('Structure', app.mol))

    sorted_result_keys = sort_model_keys(response_result_keys)

    for key in sorted_result_keys:
        # cols.append(app.Dataset.meta(key, app.cont, ((0.0, 1.0), (0, 0))))

        # createCategoryHeader creates instead of:
        # (4, '', 24, (u'BCRP_RFC', u'', u'', u'', u'', u'', u'', (0, 204, 0), [], ''), u'', u'', {0: 'yes', 1: 'no'}, (0, 0, 0, 0))
        # two tuples:
        # ((4, '', 24, (u'BCRP_RFC', u'', u'', u'', u'', u'', u'', (0, 204, 0), [], ''), u'', u'', {0: 'yes', 1: 'no'}, (0, 0, 0, 0)), {0: 'yes', 1: 'no'})
        # Which leads to problems when certain functions (setColumnColor, getColumnType, ...) are called
        # To overcome this, just add first tuple to columns
        cat_header = app.Dataset.createCategoryHeader(key, ["yes", "no"])
        cols.append(cat_header[0])

    # Insert the result if prediction is in domain or not
    domain_list = []
    for i in range(0, len(response_result_domain)):
        domain_list.append(str(
            response_result_domain[i]['model'] + " - " + response_result_domain[i]['method']))

    domain_list = sorted(domain_list)

    for i in range(0, len(domain_list)):
        cols.append(app.Dataset.meta(domain_list[i]))
    return cols


def single_prediction(appui):
    """
    Receive a compound as input and perform a prediction with the chosen models
    :param appui:
    """
    name = "Stardrop prediction"

    try:
        web_service = _create_web_service(appui)

        if web_service is not None:
            smiles = ""
            while smiles == "":
                ms = PythonQt.MoleculeSketcher()
                if ms.run():
                    mol = str(ms.getMoleculeAsMolfile())
                    smiles = ms.getMoleculeAsSMILES()
                    if smiles == "":
                        appui.show("Please insert a structure")
                    else:
                        try:
                            model_names, target_names = _get_models(web_service)
                            model_names['Select all'] = 'Select all'
                            # model_names = _get_models_gene(web_service)
                        except Exception as e:
                            log.log("There was an unexpected error: " + str(e))
                            appui.show("There was an unexpected error: " + str(e))

                        os = PythonQt.OptionSelector(False)
                        os.setWindowTitle("Select Models")
                        os.setLabel("Select one or more models")
                        sorted_models = sorted(model_names)
                        os.setAvailableItems(sorted_models)
                        models = []

                        '''
                        models = []
                        st = PythonQt.SimpleTree()
                        st.setWindowTitle("Select Model")
                        st.setLabel("Select one or more models")
                        options = {}
                        options["Support Vector Classifier (SVC)"] = model_names
                        options["Random Forest Classifier (RFC)"] = model_names
                        options["k-Nearest Neighbors Classifier (kNNC)"] = model_names
                        st.setAvailableReports(options)
                        if st.run():  # True if user clicked OK rather than Cancel
                            models = st.getSelectedReport()
                        '''

                        # Iterates as long as no model is chosen or single prediction is cancelled.
                        while len(models) == 0:
                            # Retrieve list of available models from Transportermodels and display them in StarDrop
                            if os.run():  # True if user clicked OK rather than Cancel
                                models = list(os.getSelectedItems())
                                # Clean list really necessary or is unicode list enough?
                                if 'Select all' in models:
                                    sorted_models.remove('Select all')
                                    models = sorted_models
                                if len(models) == 0:
                                    appui.show("Please choose at least one model")
                                else:
                                    clean_list = []  # Consists of of models without unicode u'
                                    for model in models:
                                        clean_list.append(str(model_names[model]))
                                    models = clean_list

                                    # Progress bar
                                    prog = PythonQt.SimpleProgress()
                                    prog.setWindowTitle("Single Prediction")
                                    prog.setMessage("Make prediction of compound")
                                    prog.show()
                                    prog.setProgress(5)

                                    # Request for single prediction
                                    response, status = web_service.post_quickprediction(name, models, mol)

                                    if status == 200:
                                        response_result = json.loads(response.data.decode('utf-8'))
                                        if response_result['error'] == "Successful":
                                            response_result_keys = _get_response_result_keys(response_result)
                                            response_result_domain = list(
                                                response_result[
                                                    'ad'])  # contains values if prediction is in domain or not

                                            try:
                                                dataset = _create_single_prediction_dataset(appui, smiles,
                                                                                            response_result,
                                                                                            response_result_keys,
                                                                                            response_result_domain,
                                                                                            model_names, prog)
                                            except Exception as e:
                                                prog.hide()
                                                log.log("Error while altering dataset: " + str(e))
                                                appui.show("Error while altering dataset: " + str(e))
                                                break

                                            if dataset is not None and dataset.cols[0]:
                                                try:
                                                    prog.setProgress(99)

                                                    appui.exportDS("Single prediction", dataset.pack())
                                                    appui.showDS("Single prediction")
                                                except Exception as e:
                                                    prog.hide()
                                                    log.log("There was an unexpected error: " + str(e))
                                                    appui.show("There was an unexpected error: " + str(e))
                                        else:
                                            prog.hide()
                                            log.log("Molecule could not be standardised.")
                                            appui.show("Molecule could not be standardised.")
                                            break
                                    else:
                                        prog.hide()
                                        log.log("There was an error getting the prediction result: " + str(
                                            response.status) + " " + response.reason)
                                        appui.show("There was an error getting the prediction result: " + str(
                                            response.status) + " " + response.reason)
                                        break
                            else:
                                break
                else:
                    break
    except Exception as e:
        log.log("Error while getting connection with Transporter Models: " + str(e))
        appui.show("Error while getting connection with Transporter Models")


def single_prediction_selected_model(appui):
    name = "Stardrop prediction"
    try:
        web_service = _create_web_service(appui)
        imported_data = appui.importDS("")
        dataset = app.Dataset()
        dataset.parse(imported_data)
        _check_dataset_input(dataset)
        if web_service is not None:
            selected_molecule = appui.getCurrentMolecule()
            selected_row = appui.getSelectedRows("")
            mol = appui.getMolFileFromSmiles(selected_molecule)
            if selected_molecule and selected_row:
                try:
                    model_names, target_names = _get_models(web_service)
                    model_names['Select all'] = 'Select all'
                except Exception as e:
                    log.log("There was an unexpected error: " + str(e))
                    appui.show("There was an unexpected error: " + str(e))

                os = PythonQt.OptionSelector(False)
                os.setWindowTitle("Select Models")
                os.setLabel("Select one or more models")
                sorted_models = sorted(model_names)
                os.setAvailableItems(sorted_models)
                models = []

                # Iterates as long as no model is chosen or single prediction is cancelled.
                while len(models) == 0:
                    # Retrieve list of available models from Transportermodels and display them in StarDrop
                    if os.run():  # True if user clicked OK rather than Cancel
                        models = list(os.getSelectedItems())
                        # Clean list really necessary or is unicode list enough?
                        if 'Select all' in models:
                            sorted_models.remove('Select all')
                            models = sorted_models
                        if len(models) == 0:
                            appui.show("Please choose at least one model")
                        else:
                            clean_list = []  # Consists of of models without unicode u'
                            for model in models:
                                clean_list.append(str(model_names[model]))
                            models = clean_list

                            # Progress bar
                            prog = PythonQt.SimpleProgress()
                            prog.setWindowTitle("Single Prediction")
                            prog.setMessage("Make prediction of compound")
                            prog.show()
                            prog.setProgress(5)

                            # Request for single prediction
                            response, status = web_service.post_quickprediction(name, models, mol)

                            if status == 200:
                                response_result = json.loads(response.data.decode('utf-8'))
                                if response_result['error'] == "Successful":
                                    response_result_keys = _get_response_result_keys(response_result)
                                    response_result_domain = list(
                                        response_result['ad'])  # contains values if prediction is in domain or not

                                    prog_value = 40
                                    new_cols = _create_single_molecule_prediction_columns(response_result_keys,
                                                                                          response_result_domain)
                                    if new_cols[0]:
                                        try:
                                            prog.setProgress(99)
                                            num_cols_dataset = dataset.columnSz()
                                            num_cols_new_data = len(new_cols)
                                            # Iterates over new results and insert them in dataset
                                            for i in range(0, num_cols_new_data):
                                                # Needs to be imported every loop cycle to be up-to-date with
                                                # changes
                                                imported_data = appui.importDS("")
                                                dataset = app.Dataset()
                                                dataset.parse(imported_data)

                                                """
                                                column_name = None
                                                if len(new_cols) > 2:
                                                    column_name = new_cols[i][3][0]
                                                else:
                                                    column_name = new_cols[i][0][3][0]
                                                """
                                                column_name = new_cols[i][3][0]
                                                idx = _get_index_of_name_in_dataset(dataset, column_name)
                                                if idx is not None:
                                                    if " - LOF Applicability Domain" in column_name:
                                                        value = _get_domain_value(response_result_domain, column_name)
                                                        constructed_value = app.Dataset.string(str(value))
                                                        dataset.set(idx, selected_row[0], constructed_value)
                                                        model_name, app_domain = column_name.split(" - ")
                                                        dataset.setColumnColor(idx, model_colors[model_name])
                                                    else:
                                                        value = response_result['res'][column_name[:column_name.index("_")]]['RFC']
                                                        constructed_value = None
                                                        if value >= 0.5:
                                                            constructed_value = app.Dataset.cat("yes", {
                                                                "yes": value, "no": 1 - value})
                                                        else:
                                                            constructed_value = app.Dataset.cat("no", {
                                                                "yes": value, "no": 1 - value})
                                                        dataset.set(idx, selected_row[0], constructed_value)
                                                        dataset.setColumnColor(idx, model_colors[ column_name[:column_name.index("_")]])


                                                    new_col = dataset.packColumn(idx)
                                                    appui.removeColumnFromDataSet("", idx)
                                                    appui.addColumnToDataSet("", idx, new_col)
                                                else:
                                                    if " - LOF Applicability Domain" in column_name:
                                                        # TODO Add invalid entries and real value
                                                        value = _get_domain_value(response_result_domain, column_name)
                                                        filled_col = _create_single_prediction_column_entries(dataset.rowSz(), selected_row[0], value, True)
                                                        dataset.insertColumn(new_cols[i], filled_col, num_cols_dataset)
                                                        model_name, app_domain = column_name.split(" - ")
                                                        dataset.setColumnColor(num_cols_dataset, model_colors[model_name])
                                                        new_col = dataset.packColumn(num_cols_dataset)
                                                        appui.addColumnToDataSet("", num_cols_dataset, new_col)
                                                        num_cols_dataset += 1
                                                    else:
                                                        idx = _get_index_of_name_in_dataset(dataset, " - LOF Applicability Domain")
                                                        value = response_result['res'][column_name[:column_name.index("_")]]['RFC']
                                                        filled_col = _create_single_prediction_column_entries(dataset.rowSz(), selected_row[0], value, False)
                                                        if idx is not None:
                                                            dataset.insertColumn(new_cols[i], filled_col, idx)
                                                            dataset.setColumnColor(idx, model_colors[column_name[:column_name.index("_")]])
                                                            new_col = dataset.packColumn(idx)
                                                            appui.addColumnToDataSet("", idx, new_col)
                                                            num_cols_dataset += 1
                                                        else:
                                                            # TODO Add invalid entries and real value
                                                            dataset.insertColumn(new_cols[i], filled_col,
                                                                                 num_cols_dataset)
                                                            dataset.setColumnColor(num_cols_dataset ,model_colors[column_name[:column_name.index("_")]])
                                                            new_col = dataset.packColumn(num_cols_dataset)
                                                            appui.addColumnToDataSet("", num_cols_dataset, new_col)
                                                            num_cols_dataset += 1
                                        except Exception as e:
                                            prog.hide()
                                            raise Exception("There was an error loading the predicted values: " + str(e))
                                else:
                                    prog.hide()
                                    log.log("Molecule could not be standardised.")
                                    appui.show("Molecule could not be standardised.")
                                    break
                            else:
                                prog.hide()
                                log.log("There was an error getting the prediction result: " + str(
                                    response.status) + " " + response.reason)
                                appui.show("There was an error getting the prediction result: " + str(
                                    response.status) + " " + response.reason)
                                break
                    else:
                        break
            else:
                appui.show("Please select a molecule to predict.")

    except Exception as e:
        log.log("Error while getting connection with Transporter Models: " + str(e))
        appui.show("Error while getting connection with Transporter Models.")


def prediction(appui):
    try:
        web_service = _create_web_service(appui)

        if web_service is not None:
            imported_data = appui.importDS("")
            dataset = app.Dataset()
            dataset.parse(imported_data)
            _check_dataset_input(dataset)
            sdf_file = _convert_dataset_to_sdf(dataset, appui)
            if sdf_file is None:
                appui.show("No data to predict.")
            else:
                try:
                    model_names, target_names = _get_models(web_service)
                    model_names['Select all'] = 'Select all'
                except Exception as e:
                    appui.show("There was an unexpected error: " + str(e))

                # Choose models for prediction
                os = PythonQt.OptionSelector(False)
                os.setWindowTitle("Select Models")
                os.setLabel("Select one or more models")
                sorted_models = sorted(model_names)
                os.setAvailableItems(sorted_models)
                models = []

                while len(models) == 0:
                    if os.run():
                        try:
                            models = list(os.getSelectedItems())
                            if 'Select all' in models:
                                sorted_models.remove('Select all')
                                models = sorted_models
                            if len(models) == 0:
                                appui.show("Please choose at least one model")
                            elif dataset.rowSz() >= 10 or len(models) >= 5:
                                w = PythonQt.Warning()
                                w.setWindowTitle("Warning!")
                                w.setMessage("This process may take several minutes. Do you want to proceed?")
                                if w.run():
                                    _predict_sdf(appui, web_service, models, dataset, sdf_file, model_names)
                                else:
                                    break
                            else:
                                _predict_sdf(appui, web_service, models, dataset, sdf_file, model_names)
                        except Exception as e:
                            log.log(str(e))
                            appui.show(str(e))
                    else:
                        break
    except ValueError:
        log.log("Error while processing dataset: Invalid molecule.")
        appui.show("Error while processing dataset: Invalid molecule.")
    except Exception as e:
        log.log("Error while getting connection with Transporter Models: " + str(e))
        appui.show("Error while getting connection with Transporter Models.")


def _predict_sdf(appui, web_service, models, dataset, sdf_file, model_names):
    """
    Prediction of all compounds in input file.
    :param appui:
    :param web_service: Object to connect to Transporter Models and use functionality.
    :param models: Models chosen by user to predict compounds.
    :param dataset: Dataset from Stardrop which should be predicted.
    :param sdf_file: Converted Dataset
    :param model_names: Names of the models available in Transporter Models
    """
    name = "Stardrop SDF upload prediction"
    counter = 0
    id_response = None
    status = 0
    limit = dataset.rowSz() * 20  # Limit depends on number of compounds in sdf file * 20 (Seconds a compound need to
    # be predicted in average)
    # Clean list really necessary or is unicode list enough?
    clean_list = []  # Consists of of models without unicode u'
    for model in models:
        clean_list.append(str(model_names[model]))
    models = clean_list

    # Progress bar
    prog = PythonQt.SimpleProgress()
    prog.setWindowTitle("SDF File Prediction")
    prog.setMessage("Upload SDF file and make prediction")
    prog.show()
    prog.setProgress(5)
    # Request for sdf prediction
    status = web_service.post_upload_sdf(name, models, sdf_file)
    if status == 200:
        # Ugly way to deal with computing time of compounds in Transportermodels
        time.sleep(1)  # To overcome time that is needed to upload sdf and appears in prediciton

        response, status = web_service.get_predictions()
        if status == 200:
            result = json.loads(response.data.decode('utf-8'))
            response_result_ids = list(result['predictions'])
            # Get list of id's from predictions
            id_response_list = []
            prog.setProgress(10)

            # Find the latest prediction id to show the result in Stardrop
            latest_id = response_result_ids[0]['id']
            for index in range(1, len(response_result_ids)):
                if response_result_ids[index]['id'] > latest_id:
                    latest_id = int(response_result_ids[index]['id'])

            # Repeatedly tries to get result either response arrives or counter of tries
            # reaches limit.
            while counter < limit:
                id_response, status = web_service.get_prediction_with_id(str(latest_id))
                if status != 200:
                    counter += 1
                else:
                    break
                time.sleep(1)
            prog.setProgress(40)

            if status == 200:
                # Get data of prediction
                id_result = json.loads(id_response.data.decode('utf-8'))

                # If not all compounds are predicted, columns can't be appended to current
                # dataset because it seems like if the number of rows from Transporter Models
                # are different to the number of rows in dataset, nothing happens after
                # appui.addColumnToDataSet()
                if len(id_result['result']) != dataset.rowSz():
                    prog.hide()
                    # raise RunTimeError("There was a problem predicting all compounds in Transporter Models.") from e
                    raise Exception("There was a problem predicting all compounds in Transporter Models.")

                response_result_keys = _get_response_result_keys(id_result['result'][0])
                id_response_list.append(id_result)

                prog_value = 40
                new_ds = _create_sdfsubmit_dataset(id_response_list, id_result,
                                                   response_result_keys, prog, prog_value)

                if new_ds.cols[0]:
                    try:
                        prog.setProgress(99)
                        num_cols_dataset = dataset.columnSz()
                        num_cols_new_dataset = new_ds.columnSz()

                        # Iterates over new results and insert them in dataset
                        for i in range(0, num_cols_new_dataset):
                            # Needs to be imported every loop cycle to be up-to-date with
                            # changes
                            imported_data = appui.importDS("")
                            dataset = app.Dataset()
                            dataset.parse(imported_data)

                            new_col = new_ds.packColumn(i)
                            idx = _get_index_of_name_in_dataset(dataset, new_ds.columnName(i))
                            if idx is not None:
                                appui.removeColumnFromDataSet("", idx)
                                appui.addColumnToDataSet("", idx, new_col)
                            else:
                                if " - LOF Applicability Domain" in new_ds.columnName(i):
                                    appui.addColumnToDataSet("", num_cols_dataset, new_col)
                                    num_cols_dataset += 1
                                else:
                                    idx = _get_index_of_name_in_dataset(dataset, " - LOF Applicability Domain")
                                    if idx is not None:
                                        appui.addColumnToDataSet("", idx, new_col)
                                        num_cols_dataset += 1
                                    else:
                                        appui.addColumnToDataSet("", num_cols_dataset, new_col)
                                        num_cols_dataset += 1
                    except Exception as e:
                        prog.hide()
                        raise Exception("There was an error loading the predicted values.")
            else:
                prog.hide()
                # raise RunTimeError("There was an error getting the prediction values.") from e
                raise Exception("There was an error while uploading the SDF file.")
                # log.log("There was an error getting the prediction values.")
                # appui.show("There was an error getting the prediction values.")
        else:
            prog.hide()
            raise Exception("There was an unexpected error: " + str(response.status) + " " + response.reason)

    else:
        prog.hide()
        raise Exception("There was an error while uploading the SDF file.")


def sort_model_keys(model_names):
    """
    Sorts the model names alphabetically and by classifier (SVC, RFC, Knnc)
    :param model_names: dictionary containing the model names
    :return: sorted list of model names
    """
    sorted_models = []
    keywords = ['_SVC', '_RFC', '_kNNC']
    model_names_alphabetically = sorted(model_names)
    for keyword in keywords:
        for name in model_names_alphabetically:
            if keyword in name:
                sorted_models.append(name)

    return sorted_models


def _convert_dataset_to_sdf(dataset, appui):
    """
    Converts the data of the dataset object to an sdf file structure.
    Just Id, mol and name are converted, the rest will be ignored.
    :param dataset:
    :param appui:
    :return:
    """
    id = ""
    mol = None
    name = None
    # associated_data = [] In case we need other data too
    sdf = ""
    for row in range(0, dataset.rowSz()):
        name = None
        mol = None
        for column in range(0, dataset.columnSz()):
            if dataset.columnType(column) == 16:  # 16 is Mol
                molecule_entry = dataset.cols[column][row]
                molecule = extract_molecule_data(molecule_entry)
                mol = appui.getMolFileFromStarDropMolecule(molecule)
                # mol = appui.getMolFileFromSmiles(str(dataset.get(column, row)[0]))
                mol = mol[:len(mol) - 5]  # removes $$$$ to insert name before end of mol file
            elif dataset.columnName(column) == 'ID':
                id = str(dataset.get(column, row))
                id = id.strip("\x00")
            elif dataset.columnName(column) == 'Name':
                name = str(dataset.get(column, row))
                name = name.strip("\x00")

        if mol is not None:
            if name is not None:
                sdf = sdf + id + mol + ">  <Name>\n" + name + "\n\n$$$$\n"
            else:
                sdf = sdf + id + mol + "\n\n$$$$\n"
    if sdf == "":
        sdf = None
    return sdf


def extract_molecule_data(molecule_entry):
    return molecule_entry[2][4][1]


def _get_models_gene(web_service):
    response_data = None
    gene_names = []
    response, status = web_service.get_available_models()
    if status == 200:
        response_data = json.loads(response.data.decode('utf-8'))
        response_data = response_data['availablemodels']
        for model in response_data:
            gene_names.append(str(model['gene']))
        gene_names = list(dict.fromkeys(gene_names))
        return gene_names
    else:
        log.log("There was an unexpected error while loading the models: " + response.reason)
        raise Exception("There was an unexpected error while loading the models: " + response.reason)


def _get_models(web_service):
    """
    Return all available models which can be used for prediction
    :param web_service: Service which is connected to Transporter Models.
    """
    response_data = None
    target_names = []
    display_names = []
    # web_service = _create_web_service(appui)
    response, status = web_service.get_available_models()
    if status == 200:
        response_data = json.loads(response.data.decode('utf-8'))
        response_data = response_data['availablemodels']
        for model in response_data:
            target_names.append(str(model['targetname']))
            display_names.append(str(model['displayname']))  # Not really necessary?

        model_names = _get_model_names(target_names, display_names)
        return model_names, target_names
    else:
        log.log("There was an unexpected error while loading the models: " + response.reason)
        raise Exception("There was an unexpected error while loading the models: " + response.reason)


def _get_model_names(target_names, models):
    """
    Get the names of the models and add as sufffix the used classification method. E.g. BSEP_RFC.
    :param target_names: List of the targets with classifier as sufffix
    :param models: Model names from the Transportmodels
    :return: Returns all models names.
    """
    model_names = {}
    for i in range(0, len(target_names)):
        underscore_idx = target_names[i].find("_")
        if "RFC" in target_names[i]:
            model_name = models[i] + target_names[i][underscore_idx:len(target_names)]
            model_names[model_name] = target_names[i]
        '''
        if "KNC" not in target_names[i]:
            model_name = models[i] + target_names[i][underscore_idx:len(target_names)]
            model_names[model_name] = target_names[i]
        else:
            model_name = models[i] + "_kNNC"
            model_names[model_name] = target_names[i]
        '''
    return model_names


def _get_column_names_from_dataset(dataset):
    """
    Get all column names from a Dataset and return them as as list as app.Dataset.meta or Dataset CategoryHeader Objects
    :param dataset: Object which column names should be retrieved
    :return: columns List of column names as app.Dataset.meta Objects from Dataset
    """
    columns = []
    for column in range(0, dataset.columnSz()):
        if dataset.columnType(column) == 16:
            columns.append(app.Dataset.meta(dataset.columnName(column), app.mol))
        elif dataset.columnName(column) == "Name" or "- LOF Applicability Domain" in dataset.columnName(
                column) or dataset.columnName(column) == "ID":
            columns.append(app.Dataset.meta(dataset.columnName(column)))
        else:
            # columns.append(app.Dataset.meta(dataset.columnName(column), app.cont, ((0.0, 1.0), (0, 0))))
            cat_header = app.Dataset.createCategoryHeader(dataset.columnName(column), ["yes", "no"])
            columns.append(cat_header[0])

    return columns


def _merge_columns_in_order(old_columns, new_columns, models):
    """
    Merges the old and the new columns in order according to the models provided by the Transportmodels.
    :param old_columns: List with column names from previous "Single prediction".
    :param new_columns: List with column names from current "Single prediction".
    :param models: List with all the available models from Transportmodels.
    :return: List of all columns merged together.
    """
    model_values = []
    domains = []
    domain_names_list = []

    for i in range(0, len(models)):
        insert = False
        # If model name is in old columns insert it model_values and domains
        if models[i] in str(old_columns):
            idx = _get_index_of_name_in_columns(old_columns, models[i])
            model_values.append(old_columns[idx])
            domain_name = models[i][:models[i].index("_")]  # Get name of domain without _svc, _rfc or _knnc
            domain_idx = _get_index_of_name_in_columns(old_columns, domain_name + " - LOF Applicability Domain")
            if domain_name not in domain_names_list:
                domains.append(
                    old_columns[domain_idx])
                domain_names_list.append(domain_name)
            insert = True

        # If model name is in new columns insert it model_values and domains
        if models[i] in str(new_columns) and insert is False:
            idx = _get_index_of_name_in_columns(new_columns, models[i])
            model_values.append(new_columns[idx])
            domain_name = models[i][:models[i].index("_")]  # Get name of domain without _svc, _rfc or _knnc
            if domain_name not in domain_names_list:
                domain_idx = _get_index_of_name_in_columns(new_columns, domain_name + " - LOF Applicability Domain")
                domains.append(
                    new_columns[domain_idx])
                domain_names_list.append(domain_name)
    domains = sorted(domains, key=lambda tup: tup[3][0])  # Sorts by the name of the tuple list

    merged_columns = model_values + domains
    merged_columns.insert(0, old_columns[0])  # "Name" column
    merged_columns.insert(1, old_columns[1])  # "Structure" column
    return merged_columns


def _get_index_of_name_in_columns(columns, name):
    """
    Get index of a specific column name in a list of column names.
    :param columns: List containing the column names.
    :param name: Name which are searched for in list.
    :return: The index of the name in the list.
    """
    idx = None
    for i in range(0, len(columns)):
        if name in str(columns[i][3][0]):
            idx = i
            break
    return idx


def _get_index_of_name_in_dataset(dataset, name):
    """
    Get index of a specific column name in the dataset.
    :param dataset: Dataset which will be searched.
    :param name: Name which are searched for in the dataset.
    :return: The index of the column in the dataset
    """
    idx = None
    for i in range(0, dataset.columnSz()):
        if name in dataset.columnName(i):
            idx = i
            break
    return idx


def _fill_new_dataset_with_values(new_dataset, dataset, old_column_names, response_result, response_result_keys,
                                  response_result_domain, smiles):
    """
    Fill the new dataset with values from the old dataset and the new result from the "Single Prediction".
    :param new_dataset: Dataset which will be filled with values from the old Dataset and new predicted values.
    :param dataset: Dataset which contains the old values (old Dataset).
    :param old_column_names: Column names from the old dataset.
    :param response_result: Result from "Quick Prediction" made by Transporter Models.
    :param response_result_keys: Keys from the dictionary of the result which contains the model values.
    :param response_result_domain: Contains the results if prediction is in domain or not.
    :param smiles: Smiles code of the predicted molecule.
    """
    # Looks if new dataset has columns in common with old dataset, if yes copies all values from the old column to
    # new dataset
    for column in range(0, new_dataset.columnSz()):
        if new_dataset.columnName(column) in str(old_column_names):
            values = []
            for row in range(0, dataset.rowSz()):
                idx = dataset.columnIndex(new_dataset.columnName(column))
                if dataset.columnType(idx) == 16:  # 16 is Mol

                    values.append(app.Dataset.molecule(dataset.get(idx, row)[0]))  # [0]
                elif dataset.columnName(idx) == "Name" or "- LOF Applicability Domain" in dataset.columnName(
                        idx) or dataset.columnName(idx) == "ID":
                    if len(dataset.get(idx, row)) > 0:
                        values.append(app.Dataset.string(dataset.stringval(idx, row)))
                    else:
                        values.append(app.Dataset.createInvalidEntry(app.string))
                else:
                    if len(dataset.get(idx, row)) > 0:
                        # For some reason BCRP Model just get {'no': x} instead of {'yes': x, 'no': y}
                        # To get no error check for length of dict
                        value = None
                        value_dict = dataset.get(idx, row)[1]
                        if len(value_dict) > 1:
                            value = value_dict['yes']
                        else:
                            if 'yes' in value_dict:
                                value = value_dict['yes']
                            else:
                                value = 1 - value_dict['no']
                        if value >= 0.5:
                            values.append(app.Dataset.cat("yes", {"yes":value,
                                                                  "no": 1 - value}))
                        else:
                            values.append(app.Dataset.cat("no", {"yes": value,
                                                                 "no": 1 - value}))
                    else:
                        values.append(app.Dataset.createDefaultCategoryEntry({0: 'yes', 1: 'no'}))

            new_dataset.setColumn(column, values)
            new_dataset.setColumnColor(column, _get_column_color_by_name(new_dataset.columnName(column)))
        else:
            values = []

            for row in range(0, dataset.rowSz()):
                if "- LOF Applicability Domain" in new_dataset.columnName(column):
                    values.append(app.Dataset.createInvalidEntry(app.string))
                else:
                    # values.append(app.Dataset.createInvalidEntry(app.cont))
                    values.append(app.Dataset.createDefaultCategoryEntry({0: 'yes', 1: 'no'}))

            new_dataset.setColumn(column, values)
            new_dataset.setColumnColor(column, _get_column_color_by_name(new_dataset.columnName(column)))

    idx_list = []  # Tracks which indices have a value inside to insert empty values in those which are not in the list
    # Insert new values in new dataset
    new_dataset.cols[0].append(app.Dataset.string(str(response_result['name'])))
    new_dataset.cols[1].append(app.Dataset.molecule(str(smiles)))
    response_result_keys = sorted(response_result_keys)
    for key in response_result_keys:
        model, classifier = key.split("_")
        result_value = float(response_result['res'][str(model)][str(classifier)])
        idx = new_dataset.columnIndex(key)
        if result_value >= 0.5:
            new_dataset.cols[idx].append(app.Dataset.cat("yes", {"yes": result_value, "no": 1 - result_value}))
        else:
            new_dataset.cols[idx].append(app.Dataset.cat("no", {"yes": result_value, "no": 1 - result_value}))

        new_dataset.setColumnColor(idx, model_colors[model])
        idx_list.append(idx)

    # Insert the result if prediction is in domain or not
    for i in range(0, len(response_result_domain)):
        name_domain = str(
            response_result_domain[i]['model'] + " - " + response_result_domain[i]['method'])
        domain_value = str(response_result_domain[i]['inDomain'])
        idx = new_dataset.columnIndex(name_domain)
        new_dataset.cols[idx].append(app.Dataset.string(domain_value))
        new_dataset.setColumnColor(idx, model_colors[response_result_domain[i]['model']])
        idx_list.append(idx)
    idx_list.sort()

    # Fill remaining empty cells from new result with No value in new dataset
    for i in range(2, new_dataset.columnSz()):
        if i not in idx_list:
            if "- LOF Applicability Domain" in new_dataset.columnName(i):
                new_dataset.cols[i].append(app.Dataset.createInvalidEntry(app.string))
            else:
                # new_dataset.cols[i].append(app.Dataset.createInvalidEntry(app.cont))
                new_dataset.cols[i].append(app.Dataset.createDefaultCategoryEntry({0: 'yes', 1: 'no'}))


def _create_single_prediction_dataset(appui, smiles, response_result, response_result_keys, response_result_domain,
                                      model_names, prog):
    """
    Creates or alters the dataset for single prediction.
    :param appui: Object needed for Stardrop.
    :param smiles: SMILES of the compound which will be predicted.
    :param response_result: Result from "Quick Prediction" made by Transporter Models.
    :param response_result_keys: Keys from the dictionary of the result which contains the model values.
    :param response_result_domain: Contains the results if prediction is in domain or not.
    :param model_names: Names of all models available for single prediction.
    :param prog: Progress bar.
    :return: Dataset filled with new values from prediction or None if existing Dataset is processed.
    """
    # Here the old dataset is alters because sending another dataset back would open a new tab in Stardrop with the made prediction
    prog.setProgress(10)
    dataset_names = appui.getDataSetNames()
    dataset = None
    if "Single prediction" in str(dataset_names):
        imported_data = appui.importDS("Single prediction")

        # Alter current dataset and send it to Stardrop
        try:
            dataset = app.Dataset()
            dataset.parse(imported_data)
            # Will later be changed including new column names
            current_column_names = _get_column_names_from_dataset(dataset)
            old_column_names = _get_column_names_from_dataset(dataset)
            new_column_names = _create_single_prediction_columns(response_result_keys,
                                                                 response_result_domain)

            sorted_model_name_keys = sort_model_keys(model_names.keys())
            merged_columns = _merge_columns_in_order(current_column_names,
                                                     new_column_names,
                                                     sorted_model_name_keys)
            new_dataset = app.Dataset(merged_columns)  # insert the updated column names
            prog.setProgress(50)

            _fill_new_dataset_with_values(new_dataset, dataset, old_column_names,
                                          response_result, response_result_keys,
                                          response_result_domain, smiles)

            # To add the new columns to the current Dataset, a dummy row has to be added,
            # otherwise it will fail.
            old_dataset_row = dataset.packRow(dataset.rowSz() - 1)
            appui.addRowToDataSet("Single prediction", dataset.rowSz(), old_dataset_row)
            prog.setProgress(70)

            # Appending new columns to Dataset
            for i in range(0, new_dataset.columnSz()):
                new_dataset_column = new_dataset.packColumn(i)
                appui.addColumnToDataSet("Single prediction", dataset.columnSz() + i,
                                         new_dataset_column)
            prog.setProgress(80)
            # Removing all column entries before the newly appended ones
            for i in range(dataset.columnSz() - 1, -1, -1):
                appui.removeColumnFromDataSet("Single prediction", i)

        except Exception as e:
            raise Exception(e)
        # Return None to signal that dataset was loaded and already processed
        # Sending another dataset back would open a new tab in Stardrop with the made prediction
        return None
    else:
        cols = _create_single_prediction_columns(response_result_keys,response_result_domain)
        prog.setProgress(50)
        if response_result:
            dataset = app.Dataset(cols)
            dataset.cols[0].append(app.Dataset.string(str(response_result['name'])))
            dataset.setColumnColor(0, [0, 0, 0])
            dataset.cols[1].append(app.Dataset.molecule(str(smiles)))
            dataset.setColumnColor(1, [0, 0, 0])

            # Insert the prediction values of the models
            for key in response_result_keys:
                model, classifier = key.split("_")
                value = float(response_result['res'][str(model)][str(classifier)])
                key_index = _get_index_of_name_in_columns(cols, key)

                if value >= 0.5:
                    dataset.cols[key_index].append(app.Dataset.cat("yes", {"yes": value, "no": (1 - value)}))
                else:
                    dataset.cols[key_index].append(app.Dataset.cat("no", {"yes": value, "no": (1 - value)}))

                dataset.setColumnColor(key_index, model_colors[model])

            # Insert the result if prediction is in domain or not
            for i in range(0, len(response_result_domain)):
                domain_value = str(response_result_domain[i]['inDomain'])
                domain_key_index = _get_index_of_name_in_columns(cols, response_result_domain[i][
                    'model'] + ' - LOF Applicability Domain')
                dataset.cols[domain_key_index].append(app.Dataset.string(domain_value))
                dataset.setColumnColor(domain_key_index, model_colors[response_result_domain[i]['model']])
            prog.setProgress(80)
        else:
            prog.hide()
            appui.show("No result found")

        return dataset


def _create_sdfsubmit_dataset_old(id_response_list, id_result, response_result_keys, prog, prog_value):
    """
    Creates the dataset for the Prediction.
    :param id_response_list: List of Id's from predictions.
    :param id_result: Containing the result of each prediction corresponding to specific id.
    :param response_result_keys: Keys from the dictionary of the result which contains the model values.
    :param prog_value: Current status of progress bar.
    :param prog_value: Current status of progress bar.
    :return: Dataset filled with all values from all predictions of the sdf file the user made with Transporter Models.
    """
    prog.setProgress(prog_value)

    # Sets the increment value for each interation for the progress bar
    increment = None
    if len(id_result['result']) == 0:
        increment = 50 / 1
    else:
        increment = 50 / len(id_result['result'])

    cols = _create_prediction_columns(id_response_list)
    dataset = app.Dataset(cols)

    for i in range(0, len(id_result['result'])):
        prog_value += increment
        prog.setProgress(prog_value)
        prog.setProgress(prog_value)

        result = id_result['result'][i]
        domain_model_index = _get_indices_of_domain_models(result)
        dataset.cols[0].append(app.Dataset.string(str(result['name'])))
        dataset.setColumnColor(0, (0, 0, 0))
        dataset.cols[1].append(
            app.Dataset.string('Prediction of molecules from SDF-file'))
        dataset.setColumnColor(1, (0, 0, 0))
        dataset.cols[2].append(app.Dataset.molecule(str(result['smiles'])))
        dataset.setColumnColor(2, (0, 0, 0))

        # Insert the prediction values of the models
        for key in response_result_keys:
            model, classifier = key.split("_")
            result_value = float(result['res'][str(model)][str(classifier)])
            idx = dataset.columnIndex(key)
            if result_value >= 0.5:
                dataset.cols[idx].append(app.Dataset.cat("yes", {"yes": result_value, "no": 1 - result_value}))
            else:
                dataset.cols[idx].append(app.Dataset.cat("no", {"yes": result_value, "no": 1 - result_value}))
            dataset.setColumnColor(idx, model_colors[model])

        # Insert the result if prediction is in domain or not
        domain_result = result['ad']
        for j in range(0, len(domain_result)):
            domain_name = str(domain_result[j]['model'] + " - " + domain_result[j]['method'])
            domain_value = str(domain_result[j]['inDomain'])
            idx = dataset.columnIndex(domain_name)
            dataset.cols[idx].append(app.Dataset.string(domain_value))
            dataset.setColumnColor(idx, model_colors[domain_result[j]['model']])

    return dataset


def _create_single_selected_molecule_dataset(response_result, response_result_keys, response_result_domain, prog,
                                             prog_value):
    """
    Creates dataset for a single selected molecule computed by TransporterModels.
    :param response_result:
    :param response_result_keys:
    :param response_result_domain:
    :param prog:
    :param prog_value:
    """
    prog.setProgress(prog_value)

    # Sets the increment value for each interaction for the progress bar
    increment = 50
    cols = _create_single_molecule_prediction_columns(response_result_keys, response_result_domain)

    dataset = app.Dataset(cols)

    prog_value += increment
    prog.setProgress(prog_value)
    # prog.setProgress(prog_value)

    # Insert the prediction values of the models
    for key in response_result_keys:
        model, classifier = key.split("_")
        result_value = float(response_result['res'][str(model)][str(classifier)])
        idx = dataset.columnIndex(key)
        if result_value >= 0.5:
            dataset.cols[idx].append(app.Dataset.cat("yes", {"yes": result_value, "no": 1 - result_value}))
        else:
            dataset.cols[idx].append(app.Dataset.cat("no", {"yes": result_value, "no": 1 - result_value}))
        dataset.setColumnColor(idx, model_colors[model])

    # Insert the result if prediction is in domain or not
    # domain_result = response_result['ad']
    for j in range(0, len(response_result_domain)):
        domain_name = str(response_result_domain[j]['model'] + " - " + response_result_domain[j]['method'])
        domain_value = str(response_result_domain[j]['inDomain'])
        idx = dataset.columnIndex(domain_name)
        dataset.cols[idx].append(app.Dataset.string(domain_value))
        dataset.setColumnColor(idx, model_colors[response_result_domain[j]['model']])

    return dataset


def _create_sdfsubmit_dataset(id_response_list, id_result, response_result_keys, prog, prog_value):
    """
    Creates dataset with computed columns from Transporter Models for Prediction.
    :param id_response_list: List of Id's from predictions.
    :param id_result: Containing the result of each prediction corresponding to specific id.
    :param response_result_keys: Keys from the dictionary of the result which contains the model values.
    :param prog_value: Current status of progress bar.
    :param prog_value: Current status of progress bar.
    :return: Dataset filled with all values from all predictions of the sdf file the user made with Transporter Models.
    """
    prog.setProgress(prog_value)

    # Sets the increment value for each interaction for the progress bar
    increment = None
    if len(id_result['result']) == 0:
        increment = 50 / 1
    else:
        increment = 50 / len(id_result['result'])
    cols = _create_prediction_columns(id_response_list)

    dataset = app.Dataset(cols)

    for i in range(0, len(id_result['result'])):
        prog_value += increment
        prog.setProgress(prog_value)
        # prog.setProgress(prog_value)

        result = id_result['result'][i]
        domain_model_index = _get_indices_of_domain_models(result)

        # Insert the prediction values of the models
        for key in response_result_keys:
            model, classifier = key.split("_")
            result_value = float(result['res'][str(model)][str(classifier)])
            idx = dataset.columnIndex(key)
            # dataset.cols[idx].append(app.Dataset.cont(result_value, 0.0))
            if result_value >= 0.5:
                dataset.cols[idx].append(app.Dataset.cat("yes", {"yes": result_value, "no": 1-result_value}))
            else:
                dataset.cols[idx].append(app.Dataset.cat("no", {"yes": result_value, "no": 1 - result_value}))
            dataset.setColumnColor(idx, model_colors[model])

        # Insert the result if prediction is in domain or not
        domain_result = result['ad']
        for j in range(0, len(domain_result)):
            domain_name = str(domain_result[j]['model'] + " - " + domain_result[j]['method'])
            domain_value = str(domain_result[j]['inDomain'])
            idx = dataset.columnIndex(domain_name)
            dataset.cols[idx].append(app.Dataset.string(domain_value))
            dataset.setColumnColor(idx, model_colors[domain_result[j]['model']])

    return dataset


def _get_indices_of_domain_models(result):
    """

    :param result:
    :return:
    """
    model_dict = {}
    for i in range(0, len(result['ad'])):
        domain_name = str(result['ad'][i]['model'])
        domain_value = str(result['ad'][i]['inDomain'])
        model_dict[domain_name] = domain_value
    return model_dict


def _get_response_result_keys(response_result):
    """
    Creates a list with the names of the models with their classifier combined. Like BSEP_SVC or MATE1_kNNC.
    :param response_result:
    :return:
    """
    keys = list(response_result['res'].keys())
    full_key_names = []
    for key in keys:
        classifiers = response_result['res'][key].keys()
        for classifier in classifiers:
            full_key_names.append(key + "_" + classifier)
    return full_key_names


def _get_column_color_by_name(name):
    """
    Get color of column by name.
    :param name:
    :return: Returns rgb code of the color from the column name
    """
    # Sets the right column color for each column
    color = ()
    if "_" in name:
        model, classifier = name.split("_")
        color = model_colors[model]
    elif " - " in name:
        model, classifier = name.split(" - ")
        color = model_colors[model]
    else:
        color = (0, 0, 0)
    return color


def _update_response(dataset, id_response_list):
    """
    Checks if in the response of Transporter Models are results which are also in the current dataset and removes them.
    :param response_list:
    """
    # TODO uncomplete, only filter - Lof and not _RFC and - Lof
    dataset_columns = _get_column_names_from_dataset(dataset)
    dataset_column_names = list()
    for i in range(0, len(dataset_columns)):
        column_name = None
        column_name = dataset_columns[i][3][0]
        """
        if len(dataset_columns) > 2:
            column_name = dataset_columns[i][3][0]
        else:
            column_name = dataset_columns[i][0][3][0]
        """
        dataset_column_names.append(column_name)
    for i in range(0, len(id_response_list)):
        for j in range(0, len(id_response_list[i]['result'])):
            for k in range(0, len(list(id_response_list[i]['result'][j]['ad']))):
                name = str(id_response_list[i]['result'][j]['ad'][k]['model'])
                if name + "_RFC" in dataset_column_names or name + " - LOF Applicability Domain" in dataset_column_names:
                    del id_response_list[i]['result'][j]['ad'][k]


def _check_dataset_input(dataset):
    """
    Checks dataset for invalid molecule. If invalid value is found, ValueError is raised, otherwise nothing happens.
    :param dataset: Dataset which is checked.
    """
    for i in range(0, dataset.columnSz()):
        if dataset.columnType(i) == 16:  # 16 is mol type
            for j in range(0, dataset.rowSz()):
                if dataset.stringval(i, j) == "":
                    raise ValueError()


def _get_domain_value(response_result_domain, column_name):
    """
    Gets the the value of a certain molecule if it is in domain or not.
    The value is extracted from the response result by the model name.
    :param response_result_domain:
    :param column_name: Name of the column and also of the model used for prediction.
    :return:
    """
    name = column_name[0:column_name.index("-") - 1]
    for domain in response_result_domain:
        if name in domain['model']:
            return domain['inDomain']

    return None


def _create_single_prediction_column_entries(number_rows_dataset, selected_row, value, domain):
    """
    Creates a column for the single prediction including the predicted value and invalid
    entries for molecules that weren't predicted.
    :param number_rows_dataset:
    :param selected_row:
    :param value:
    :param domain: Decides if its an activity value or an applicability domain. True = applicability domain, False = activity value
    """
    entries = []
    if domain:
        for i in range(0, number_rows_dataset):
            if i != selected_row:
                entries.append(app.Dataset.createInvalidEntry(app.string))
            else:
                entries.append(app.Dataset.string(str(value)))
    else:
        for i in range(0, number_rows_dataset):
            if i != selected_row:
                # entries.append(app.Dataset.createInvalidEntry(app.cont))
                entries.append(app.Dataset.createDefaultCategoryEntry({0: 'yes', 1: 'no'}))
            else:
                if value >= 0.5:
                    entries.append(app.Dataset.cat("yes", {"yes": value, "no": 1 - value}))
                else:
                    entries.append(app.Dataset.cat("no", {"yes": value, "no": 1 - value}))

    return entries

# TODO use proper names for the scripts
def get_stardrop_definitions():
    definitions = [
        {
            'script_name': 'Phenaris/Draw molecule for single prediction...',
            'callback': single_prediction
        },
        {
            'script_name': 'Phenaris/Single prediction using selected model...',
            'callback': single_prediction_selected_model
        },
        {
            'script_name': 'Phenaris/Prediction...',
            'callback': prediction
        }
    ]

    return definitions
